#include <polylib/polylibgmp.h>

struct pip_dep_result {
    Polyhedron *M;
    Polyhedron *D;
};

Polyhedron *pip_lexminmax(Polyhedron *P, int pos, int n, int nparam, int max);
struct pip_dep_result pip_inputdep(Polyhedron *D, int dim, Matrix *M);
Polyhedron *Polyhedron_projectout(Polyhedron *P, int pos, int n, int nparam);
Polyhedron *Domain_projectout(Polyhedron *D, int pos, int n, int nparam);

Polyhedron *EPoly_Intersection(Polyhedron *A, Polyhedron *B, int d, int nparam);

/* Project out first dim variables that are assume to be
 * affine combination of the other variables.
 */
Polyhedron *Polyhedron_projectout_dep(Polyhedron *map, int dim);

Polyhedron *Domain_concat(Polyhedron *head, Polyhedron *tail);

/* Returns deepest level where first precedes second, where level
 * ranges over statement levels and loop levels, with 0 corresponding
 * to the outermost statement level, 1 to the outermost loop, etc.
 * If second precedes first at the outermost level, then the value is -1.
 * If first and second are surrounded by n common loops then the level
 * is 2*n if first precedes second inside the loop nest and 2*n-1 otherwise.
 */
typedef int (*pip_access_precedes_level)(void *first, void *second);

struct pip_rel {
    Polyhedron* rel;
    int		n_input;
    int		n_output;
    void*	data;
};

struct pip_dependence_info {
    Polyhedron *ctx;
    struct pip_rel read;
    pip_access_precedes_level	precedes_level;
    int		    n_write;
    struct pip_rel  write[0];
};

struct pip_rel *pip_lastwrite(struct pip_dependence_info *dep);

int pip_is_reordering(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx);
int pip_is_inorder_accesses(Polyhedron *dep1, Polyhedron *dep2,
			    int n_write, int n_read, Polyhedron *ctx);
int pip_has_multiplicity(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx);
int pip_is_consecutive_loop(Polyhedron *dep, int n_write, int n_read, 
			    Polyhedron *ctx);
int pip_is_consecutive_dep(Polyhedron *dep2, Polyhedron *dep1, 
			   int n_write, int n_read, Polyhedron *ctx);

Polyhedron *pip_min_ddv(Polyhedron *dep, Polyhedron *sw, Polyhedron *sr,
			int nwrite, int nread, int dim, Polyhedron *ctx);
