#ifndef YLL_H
#define YLL_H

#include <stdio.h>
#include <syck.h>

typedef enum _yll_option {
    YLL_INLINE_SEQ = 1 << 0,
    YLL_ADD_SPACE = 1 << 1,
} YllOption;

struct _yll_emitter;

typedef void (*YllNext)(struct _yll_emitter *e);

struct _yll_emitter_node {
    struct _yll_emitter_node *next;
    YllNext		      nextf;
    YllOption	     	      flags;
};

struct _yll_anchor_node {
    struct _yll_emitter_node	en;
    SYMID			anchor;
};

struct _yll_seq_node {
    struct _yll_emitter_node	en;
    int				first;
};

typedef struct _yll_emitter_node YllEmitterNode;
typedef struct _yll_anchor_node YllAnchorNode;
typedef struct _yll_seq_node YllSeqNode;

struct _yll_emitter {
    FILE	   *out;
    SYMID   	    last_anchor;
    YllEmitterNode *stack;
    int		    indent;
    YllOption	    flags;
    const char	   *transfer;
};

typedef struct _yll_emitter YllEmitter;

typedef void (*YllOutputHandler)(YllEmitter *, char *, long);

YllEmitter* yll_new_emitter();
void yll_emitter_handler(YllEmitter *e, YllOutputHandler hdlr);
void yll_free_emitter(YllEmitter *e);
SYMID yll_emitter_anchor(YllEmitter *e);
void yll_emitter_alias(YllEmitter *e, SYMID id);
void yll_emitter_set_options(YllEmitter *e, YllOption flags);
void yll_emitter_set_transfer(YllEmitter *e, const char *t);
void yll_emitter_start_seq(YllEmitter *e);
void yll_emitter_end_seq(YllEmitter *e);
void yll_emitter_start_map(YllEmitter *e);
void yll_emitter_end_map(YllEmitter *e);
void yll_emitter_write_string(YllEmitter *e, const char *str);
void yll_emitter_write_int(YllEmitter *e, int i);
void yll_emitter_write_null(YllEmitter *e);

#endif
